@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* exec.h: command execution */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2020 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,13 +41,13 @@ typedef enum sigtype_T {
 #define Exit_ASSGNERR Exit_ERROR
 #define Exit_REDIRERR Exit_ERROR
 
-extern int laststatus, savelaststatus;
+extern int laststatus, savelaststatus, exitstatus;
 extern pid_t lastasyncpid;
 extern _Bool special_builtin_executed;
 extern _Bool is_executing_auxiliary;
 
 struct execstate_T;
-extern void reset_execstate(void);
+extern void reset_execstate(_Bool reset_iteration);
 extern struct execstate_T *save_execstate(void)
     __attribute__((malloc,warn_unused_result));
 extern void restore_execstate(struct execstate_T *save)
@@ -60,22 +60,22 @@ extern _Bool need_break(void)
 struct and_or_T;
 struct embedcmd_T;
 extern void exec_and_or_lists(const struct and_or_T *a, _Bool finally_exit);
-extern pid_t fork_and_reset(pid_t pgid, _Bool fg, sigtype_T sigtype);
 extern struct xwcsbuf_T *get_xtrace_buffer(void);
+extern pid_t fork_and_reset(pid_t pgid, _Bool fg, sigtype_T sigtype);
 extern wchar_t *exec_command_substitution(const struct embedcmd_T *cmdsub)
     __attribute__((nonnull,malloc,warn_unused_result));
 extern int exec_variable_as_commands(
-	const wchar_t *varname, const char *codename)
+        const wchar_t *varname, const char *codename)
     __attribute__((nonnull));
 extern int exec_variable_as_auxiliary(
-	const wchar_t *varname, const char *codename)
+        const wchar_t *varname, const char *codename)
     __attribute__((nonnull));
 #define exec_variable_as_auxiliary_(varname) \
-	exec_variable_as_auxiliary(L varname, "$" varname)
+        exec_variable_as_auxiliary(L varname, "$" varname)
 
 #if YASH_ENABLE_LINEEDIT
 extern _Bool autoload_completion_function_file(
-	const wchar_t *filename, const wchar_t *cmdname)
+        const wchar_t *filename, const wchar_t *cmdname)
     __attribute__((nonnull(1)));
 extern _Bool call_completion_function(const wchar_t *funcname)
     __attribute__((nonnull));
@@ -94,7 +94,7 @@ extern int break_builtin(int argc, void **argv)
     __attribute__((nonnull));
 #if YASH_ENABLE_HELP
 extern const char break_help[], break_syntax[], continue_help[],
-	continue_syntax[];
+        continue_syntax[];
 #endif
 
 extern int eval_builtin(int argc, void **argv)
@@ -134,4 +134,4 @@ extern const char times_help[], times_syntax[];
 #endif /* YASH_EXEC_H */
 
 
-/* vim: set ts=8 sts=4 sw=4 noet tw=80: */
+/* vim: set ts=8 sts=4 sw=4 et tw=80: */
